/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2009
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_COMMON_PRIORITY_QUEUES_HPP__
#define YAYI_COMMON_PRIORITY_QUEUES_HPP__


/*!@file
 * Defines the priority queues
 *
 */

#include <map>
#include <deque>
#include <functional>


#include "Yayi/core/yayiCommon/common_errors.hpp"

namespace yayi
{
  
  // forward declaration
  template <class key_t, class value_t, class order_t>
  class priority_queue_t;
  
  template <class key_t, class plateau_container_t>
  class priority_queue_iterator_t;
  
  template <class key_t, class plateau_container_t>
  class priority_queue_plateau_iterator_t;
  
  
  
  
  /*!@brief Iterator on plateaus
   * This iterator iterates a generic iterator on a constant plateau. Since fewer data is handled during this iteration,
   * the iterator is faster than the generic @ref priority_queue_iterator_t iterator.
   *
   * @see priority_queue_iterator_t
   */
  template <class key_t, class plateau_container_iterator_t>
  class priority_queue_plateau_iterator_t {
    key_t key_value;
    plateau_container_iterator_t it;
    
  public: 
    typedef priority_queue_plateau_iterator_t<key_t, plateau_container_iterator_t> this_type;
    typedef typename plateau_container_iterator_t::value_type value_type;
    typedef typename plateau_container_iterator_t::reference_type reference_type;

    
    priority_queue_plateau_iterator_t(const key_t& k, const plateau_container_iterator_t& it_) : key_value(k), it(it_)
    {}
    
    //! Returns the actual height of the plateau 
    const key_t& key() const {
      return key_value;
    }
  
    //! Increments iterator
    this_type& operator++() {
      ++it;
      return *this;
    }
    
    //! Deference operator
    reference_type operator*() {
      return *it;
    }
    
    //! Compares iterators
    bool operator==(const this_type& r_) const throw() {
      return (r_.key_value == key_value) && (r_.it == it); // second can be cheched, so we can avoid first test if no chech is done (no debug ?)
    }
    
    //! Compares iterators
    bool operator!=(const this_type& r_) const throw() {
      return (r_.key_value != key_value) || (r_.it != it); // second can be cheched, so we can avoid first test if no chech is done (no debug ?)
    }
  };
  
  
  /*!@brief Iterator on the whole priority queue
   * This iterator iterates a generic iterator on the priority queue, ordered in the same order
   * as the ke elements and the insertion of value elements (FIFO style).
   * There is an other version iterating on a constant priority level (@ref priority_queue_plateau_iterator_t)
   *
   * @see priority_queue_plateau_iterator_t
   */
  template <class order_container_iterator_t, class plateau_container_iterator_t>
  class priority_queue_iterator_t {
    plateau_container_iterator_t  it, ite;
    order_container_iterator_t    itm;
    
  public: 
    typedef priority_queue_iterator_t<order_container_iterator_t, plateau_container_iterator_t> this_type;
    typedef typename plateau_container_iterator_t::value_type value_type;
    typedef typename plateau_container_iterator_t::reference_type reference_type;
    
    typedef typename order_container_iterator_t::mapped_type key_type;
    
    
    priority_queue_iterator_t(
      const order_container_iterator_t& itm_, const plateau_container_iterator_t& it_) : itm(itm_), it(it_)
    {}
    
    //! Returns the actual height of the plateau
    const key_type& key() const {
      return itm->first;
    }
    
    //! Increments iterator
    this_type& operator++() {
      ++it;
      if(it == ite) {
        ++itm;
        ite = itm->second.end();      // attention: faux
        it = itm->second.begin();
      }
      return *this;
    }
    
    //! Deference operator
    reference_type operator*() {
      return *it;
    }
    
    //! Compares iterators
    bool operator==(const this_type& r_) const throw() {
      return (r_.itm == itm) && (r_.it == it); // second can be cheched, so we can avoid first test if no chech is done (no debug ?)
    }
    
    //! Compares iterators
    bool operator!=(const this_type& r_) const throw() {
      return (r_.itm != itm) || (r_.it != it); // second can be checked, so we can avoid first test if no check is done (no debug ?)
    }
  };
  
  
  /*!@brief Generic priority queue
   *
   * @author Raffi Enficiaud
   */
  template <class key_t, class value_t, class order_t = std::less<key_t> >
  class priority_queue_t 
  {
  
    typedef std::deque<value_t> plateau_container_type;
    typedef std::deque< std::pair<key_t, value_t> > temporary_queue_type;
  
    typedef std::map<key_t, plateau_container_type, order_t> map_type;
    typedef typename map_type::iterator map_iterator_t;
    typedef typename map_type::const_iterator map_const_iterator_t;
  
    map_type              map_object;
    temporary_queue_type  temp_q;
  
    // mettre des static assert pour les types supportés
  
  
  
  public:
    
    typedef key_t     key_type;
    typedef value_t   value_type;
    typedef order_t   order_type;
  
  
    typedef priority_queue_plateau_iterator_t<key_t, typename plateau_container_type::iterator> plateau_iterator_type;
    typedef priority_queue_plateau_iterator_t<key_t, typename plateau_container_type::const_iterator> plateau_const_iterator_type;
    friend class priority_queue_plateau_iterator_t<key_t, typename plateau_container_type::iterator>;
    friend class priority_queue_plateau_iterator_t<key_t, typename plateau_container_type::const_iterator>;
    
    typedef priority_queue_iterator_t<map_iterator_t, typename plateau_container_type::iterator> iterator;
    typedef priority_queue_iterator_t<map_const_iterator_t, typename plateau_container_type::const_iterator> const_iterator;
    friend class priority_queue_iterator_t<map_iterator_t, typename plateau_container_type::iterator>;
    friend class priority_queue_iterator_t<map_const_iterator_t, typename plateau_container_type::const_iterator>;
  
    priority_queue_t() : map_object(), temp_q()
    {}
    
    priority_queue_t(const priority_queue_t& r) : map_object(r.map_object), temp_q(r.temp_q)
    {}
    
    
    
    
    // plateau iteration
    plateau_iterator_type begin_plateau(const key_type& k) {
      map_iterator_t itm = map_object.find(k);
      if(itm == map_object.end()) {
        YAYI_THROW("The key " << k << " does not exist");
      }
      return plateau_iterator_type(k, itm->second.begin());
    }
    plateau_iterator_type end_plateau(const key_type& k) {
      map_iterator_t itm = map_object.find(k);
      if(itm == map_object.end()) {
        YAYI_THROW("The key " << k << " does not exist");
      }
      return plateau_iterator_type(k, itm->second.end());
    }
    std::pair<plateau_iterator_type, plateau_iterator_type> range_plateau(const key_type& k) {
      map_iterator_t itm = map_object.find(k);
      if(itm == map_object.end()) {
        YAYI_THROW("The key " << k << " does not exist");
      }
      return std::make_pair(
        plateau_iterator_type(k, itm->second.begin()),
        plateau_iterator_type(k, itm->second.end())
        );
    }
    
    plateau_iterator_type begin_top_plateau() {
      map_iterator_t itm = map_object.begin();
      if(itm == map_object.end()) {
        YAYI_THROW("The map is empty");;
      }
      return plateau_iterator_type(itm->first, itm->second.begin());
    }
    plateau_iterator_type end_top_plateau() {
      map_iterator_t itm = map_object.begin();
      if(itm == map_object.end()) {
        YAYI_THROW("The map is empty");
      }
      return plateau_iterator_type(itm->first, itm->second.end());
    }
    std::pair<plateau_iterator_type, plateau_iterator_type> range_top_plateau() {
      if(map_object.empty()) {
        YAYI_THROW("The map is empty"); // on doit envoyer quelque chose de correct dans ce cas
      }
      
      map_iterator_t itm = map_object.begin();
      return std::make_pair(
        plateau_iterator_type(itm->first, itm->second.begin()),
        plateau_iterator_type(itm->first, itm->second.end())
        );
    }
    
    
    // plateau iteration (const)
    plateau_const_iterator_type begin_plateau(const key_type& k) const {
      map_const_iterator_t itm = map_object.find(k);
      if(itm == map_object.end()) {
        YAYI_THROW("The key " << k << " does not exist");
      }
      return plateau_const_iterator_type(k, itm->second.begin());    
    }
    plateau_const_iterator_type end_plateau(const key_type& k) const {
      map_const_iterator_t itm = map_object.find(k);
      if(itm == map_object.end()) {
        YAYI_THROW("The key " << k << " does not exist");
      }
      return plateau_const_iterator_type(k, itm->second.end());    
    }
    std::pair<plateau_const_iterator_type, plateau_const_iterator_type> range_plateau(const key_type& k) const {
      map_const_iterator_t itm = map_object.find(k);
      if(itm == map_object.end()) {
        YAYI_THROW("The key " << k << " does not exist");
      }
      return std::make_pair(
        plateau_const_iterator_type(k, itm->second.begin()),
        plateau_const_iterator_type(k, itm->second.end())
        );
    }
    
    
    plateau_const_iterator_type begin_top_plateau() const {
      map_const_iterator_t itm = map_object.begin();
      if(itm == map_object.end()) {
        YAYI_THROW("The map is empty");;
      }
      return plateau_const_iterator_type(itm->first, itm->second.begin());
    }
    plateau_const_iterator_type end_top_plateau() const {
      map_const_iterator_t itm = map_object.begin();
      if(itm == map_object.end()) {
        YAYI_THROW("The map is empty");
      }
      return plateau_const_iterator_type(itm->first, itm->second.end());
    }
    std::pair<plateau_const_iterator_type, plateau_const_iterator_type> range_top_plateau() const {
      if(map_object.empty()) {
        YAYI_THROW("The map is empty"); // on doit envoyer quelque chose de correct dans ce cas
      }
      
      map_const_iterator_t itm = map_object.begin();
      return std::make_pair(
        plateau_const_iterator_type(itm->first, itm->second.begin()),
        plateau_const_iterator_type(itm->first, itm->second.end())
        );
    }
    
    
    // Generic iteration
    iterator begin() {
      return iterator(map_object.begin(), map_object.begin()->second.begin());
    }
    
    iterator end() {
      return iterator(--(map_object.end()), map_object.begin()->second.begin());
    }
    
    
    
    // Insertion
    void insert(const key_type& k, const value_type &v) {
      map_object[k].push_back(v);
    }
  
  
  };
  
  
  
}


#endif /* YAYI_COMMON_PRIORITY_QUEUES_HPP__ */