/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2009
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_COMMON_VARIANT__HPP__
#define YAYI_COMMON_VARIANT__HPP__

/*!@file
 * This file contains the definitions for variant handling
 */

#include <Yayi/core/yayiCommon/common_types.hpp>
#include <Yayi/core/yayiCommon/include/common_types_T.hpp>
#include <Yayi/core/yayiCommon/common_pixels.hpp>
#include <Yayi/core/yayiCommon/common_pixels_T.hpp>

#include <iostream>
#include <vector>
#include <map>
#include <string>
#include <boost/utility/enable_if.hpp>

namespace yayi {

  // forward
  template <class U>  U& variant_field_from_type(variant&);
  
  //! This function cleans the content of a variant (from its runtime type)
  void cleanup_variant(variant& v);
  
  //! This function copies the content of a variant (according to its runtime type)
  void copy_variant(const variant& from, variant& to);
  
  //! This function clones a variant (copy and allocation, without destroying its content)
  void clone_variant_init(const variant& v1, variant& v2);
  
  //! Transformation utility
  template <class U>  typename boost::enable_if< type_support<U> >::type 
  transform_variant_utility(const U& u, variant &v);
  
  //! Specialization of @transform_variant_utility for vector types
  template <class U>  typename boost::enable_if< type_support< U > >::type 
  transform_variant_utility(const std::vector<U> &u, variant &v);
  
  //! Specialization of @transform_variant_utility for variant types
  //void transform_variant_utility(const variant& u, variant &v);


  
  
  

  //! Structure describing any type of pixel
  struct s_any_type
  {
    type        element_type;
    union
    {
      bool        b_value;
      yaUINT8     ui8_value;
      yaINT8      i8_value;
      yaUINT16    ui16_value;
      yaINT16     i16_value;
      yaUINT32    ui32_value;
      yaINT32     i32_value;
      
      yaUINT64    ui64_value;
      yaINT64     i64_value;
      
      yaF_simple  fs_value;
      yaF_double  fd_value;
      
      std::vector<s_any_type>         *vector_value;
      string_type                     *string_value;
      wide_string_type                *wstring_value;
      
      std::map<int, s_any_type>       *map_value;
      std::list<s_any_type>           *list_value;
      
      s_compound_pixel_t<bool,        mpl::int_<3> >* b_pixel_3;
      s_compound_pixel_t<yaUINT8,     mpl::int_<3> >* ui8_pixel_3;
      s_compound_pixel_t<yaINT8,      mpl::int_<3> >* i8_pixel_3;
      s_compound_pixel_t<yaUINT16,    mpl::int_<3> >* ui16_pixel_3;
      s_compound_pixel_t<yaINT16,     mpl::int_<3> >* i16_pixel_3;
      s_compound_pixel_t<yaUINT32,    mpl::int_<3> >* ui32_pixel_3;
      s_compound_pixel_t<yaINT32,     mpl::int_<3> >* i32_pixel_3;
      s_compound_pixel_t<yaUINT64,    mpl::int_<3> >* ui64_pixel_3;
      s_compound_pixel_t<yaINT64,     mpl::int_<3> >* i64_pixel_3;
      s_compound_pixel_t<yaF_simple,  mpl::int_<3> >* fs_pixel_3;
      s_compound_pixel_t<yaF_double,  mpl::int_<3> >* fd_pixel_3;
      
      std::complex<yaF_simple>        *fs_complex;
      std::complex<yaF_double>        *fd_complex;
      
      offset                          offset_value;
      
    } element;
    
    
    //! Default constructor
    s_any_type() : element_type(undefined_type) {
    }
    
    //! Copy constructor
    s_any_type(const s_any_type& c) {
      clone_variant_init(c, *this);
    }
    
    
    /*! Template constructor from any supported type
     * The second argument is not used (dummy). It is here to prevent the instanciation of the constructor
     * with unsupported types, such as s_any_type itself.
     */
    template <class U>
      s_any_type(const U& u, typename boost::enable_if< type_support< U > >::type* dummy = 0) : element_type(undefined_type) {
      transform_variant_utility(u, *this);
    }
    
    //! Assignment operator
    s_any_type& operator=(const s_any_type& c) {
      copy_variant(c, *this);
      return *this;
    }
    
    //! Destructor
    ~s_any_type() {
      cleanup_variant(*this);
    }
    
    
    //! Conversion for default types
    template <class U> operator U() const;
    
    //! Special implementation for complex types
    template <class U> operator std::complex<U>() const;
    
    //! Special implementation for vector types
    template <class U> operator std::vector<U>() const;
    
    //! Special implementation for pixel types
    template <class U, class dimension> 
    operator yayi::s_compound_pixel_t<U, dimension>() const;
    
    
  };
  
  
  // Default behaviour
  //! This template allows to access to a specific field of the variant from the provided type. 
  //! The mapping is exact 1 to 1.
  template <class U>  U&                  variant_field_from_type             (variant&)            {YAYI_THROW("Unsupported type");}
  template <class U>  const U&            variant_field_from_type             (const variant&)      {YAYI_THROW("Unsupported type");}
  
  
  // Scalar accessors
  template <> inline bool&                variant_field_from_type<bool>       (variant& v)          {return v.element.b_value;    }
  template <> inline yaUINT8&             variant_field_from_type<yaUINT8>    (variant& v)          {return v.element.ui8_value;  }
  template <> inline yaINT8&              variant_field_from_type<yaINT8>     (variant& v)          {return v.element.i8_value;   }
  template <> inline yaUINT16&            variant_field_from_type<yaUINT16>   (variant& v)          {return v.element.ui16_value; }
  template <> inline yaINT16&             variant_field_from_type<yaINT16>    (variant& v)          {return v.element.i16_value;  }
  template <> inline yaUINT32&            variant_field_from_type<yaUINT32>   (variant& v)          {return v.element.ui32_value; }
  template <> inline yaINT32&             variant_field_from_type<yaINT32>    (variant& v)          {return v.element.i32_value;  }
  template <> inline yaUINT64&            variant_field_from_type<yaUINT64>   (variant& v)          {return v.element.ui64_value; }
  template <> inline yaINT64&             variant_field_from_type<yaINT64>    (variant& v)          {return v.element.i64_value;  }
  template <> inline yaF_simple&          variant_field_from_type<yaF_simple> (variant& v)          {return v.element.fs_value;   }
  template <> inline yaF_double&          variant_field_from_type<yaF_double> (variant& v)          {return v.element.fd_value;   }
  
  template <> inline const bool&          variant_field_from_type<bool>       (const variant& v)    {return v.element.b_value;    }
  template <> inline const yaUINT8&       variant_field_from_type<yaUINT8>    (const variant& v)    {return v.element.ui8_value;  }
  template <> inline const yaINT8&        variant_field_from_type<yaINT8>     (const variant& v)    {return v.element.i8_value;   }
  template <> inline const yaUINT16&      variant_field_from_type<yaUINT16>   (const variant& v)    {return v.element.ui16_value; }
  template <> inline const yaINT16&       variant_field_from_type<yaINT16>    (const variant& v)    {return v.element.i16_value;  }
  template <> inline const yaUINT32&      variant_field_from_type<yaUINT32>   (const variant& v)    {return v.element.ui32_value; }
  template <> inline const yaINT32&       variant_field_from_type<yaINT32>    (const variant& v)    {return v.element.i32_value;  }
  template <> inline const yaUINT64&      variant_field_from_type<yaUINT64>   (const variant& v)    {return v.element.ui64_value; }
  template <> inline const yaINT64&       variant_field_from_type<yaINT64>    (const variant& v)    {return v.element.i64_value;  }
  template <> inline const yaF_simple&    variant_field_from_type<yaF_simple> (const variant& v)    {return v.element.fs_value;   }
  template <> inline const yaF_double&    variant_field_from_type<yaF_double> (const variant& v)    {return v.element.fd_value;   }
  
  
  
  // String accessors
  template <> inline string_type&                     variant_field_from_type<string_type>      (variant& v)    {return *v.element.string_value;   }
  template <> inline wide_string_type&                variant_field_from_type<wide_string_type> (variant& v)    {return *v.element.wstring_value;  }
  
  template <> inline const string_type&               variant_field_from_type<string_type>      (const variant& v)   {return *v.element.string_value;   }
  template <> inline const wide_string_type&          variant_field_from_type<wide_string_type> (const variant& v)   {return *v.element.wstring_value;  }
  
  // Complex accessors
  template <> inline std::complex<yaF_simple>&        variant_field_from_type< std::complex<yaF_simple> > (variant& v)          {return *v.element.fs_complex;   }
  template <> inline std::complex<yaF_double>&        variant_field_from_type< std::complex<yaF_double> > (variant& v)          {return *v.element.fd_complex;   }
  
  template <> inline const std::complex<yaF_simple>&  variant_field_from_type< std::complex<yaF_simple> > (const variant& v)    {return *v.element.fs_complex;   }
  template <> inline const std::complex<yaF_double>&  variant_field_from_type< std::complex<yaF_double> > (const variant& v)    {return *v.element.fd_complex;   }

  // Vector
  template <> inline std::vector<s_any_type>&         variant_field_from_type< std::vector<s_any_type> > (variant& v)           {return *v.element.vector_value;   }
  template <> inline const std::vector<s_any_type>&   variant_field_from_type< std::vector<s_any_type> > (const variant& v)     {return *v.element.vector_value;   }
  

  
  
  // Pixel 3 channels accessor
  template <> inline s_compound_pixel_t<bool,       mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<bool,       mpl::int_<3> > > (variant& v)   {return *v.element.b_pixel_3;    }
  template <> inline s_compound_pixel_t<yaUINT8,    mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaUINT8,    mpl::int_<3> > > (variant& v)   {return *v.element.ui8_pixel_3;  }
  template <> inline s_compound_pixel_t<yaINT8,     mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaINT8,     mpl::int_<3> > > (variant& v)   {return *v.element.i8_pixel_3;   }
  template <> inline s_compound_pixel_t<yaUINT16,   mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaUINT16,   mpl::int_<3> > > (variant& v)   {return *v.element.ui16_pixel_3; }
  template <> inline s_compound_pixel_t<yaINT16,    mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaINT16,    mpl::int_<3> > > (variant& v)   {return *v.element.i16_pixel_3;  }
  template <> inline s_compound_pixel_t<yaUINT32,   mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaUINT32,   mpl::int_<3> > > (variant& v)   {return *v.element.ui32_pixel_3; }
  template <> inline s_compound_pixel_t<yaINT32,    mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaINT32,    mpl::int_<3> > > (variant& v)   {return *v.element.i32_pixel_3;  }
  template <> inline s_compound_pixel_t<yaUINT64,   mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaUINT64,   mpl::int_<3> > > (variant& v)   {return *v.element.ui64_pixel_3; }
  template <> inline s_compound_pixel_t<yaINT64,    mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaINT64,    mpl::int_<3> > > (variant& v)   {return *v.element.i64_pixel_3;  }
  template <> inline s_compound_pixel_t<yaF_simple, mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaF_simple, mpl::int_<3> > > (variant& v)   {return *v.element.fs_pixel_3;   }
  template <> inline s_compound_pixel_t<yaF_double, mpl::int_<3> >&  variant_field_from_type<s_compound_pixel_t<yaF_double, mpl::int_<3> > > (variant& v)   {return *v.element.fd_pixel_3;   }
  
  
  
  


  template <class U>
  inline void clean_stack_element(variant& v) {
    U* u = &variant_field_from_type<U>(v);
    delete u;
    u = 0;
    v.element_type.c_type = type::c_unknown;
    v.element_type.s_type = type::s_undefined;
  }
         
  inline void cleanup_variant(variant& v) {
    //! TODO a completer
    switch(v.element_type.c_type){
      case type::c_unknown:
      case type::c_scalar:
        break;
      case type::c_vector:
        clean_stack_element< std::vector<s_any_type> >(v);
        break;
        
      default:
        YAYI_THROW("Undefined type " + string_type(v.element_type));                     break;
    }
  }
  


  //! Checks if the element of the variant is of the same type
  inline bool variants_same_type(const variant& v1, const variant& v2) {
    return v1.element_type == v2.element_type;
  }
  
  inline bool variants_same_type(const variant& v1, const type& t) {
    return v1.element_type == t;
  }
  
  inline bool variants_same_type(const variant& v1, const type::e_scalar_type& s, const type::e_compound_type& c) {
    return v1.element_type.s_type == s && v1.element_type.c_type == c;
  }
  
  inline void copy_variant(const variant& v1, variant& v2) {
    // Simple cases before
    if(v1.element_type.c_type == type::c_scalar)  {
       if((v2.element_type.c_type != type::c_scalar) && (v2.element_type.c_type != type::c_unknown) ) {
         cleanup_variant(v2);
         // v2.element_type = v1.element_type;
       }
       // TODO faire un test unitaire sur ce bug
       v2.element_type = v1.element_type;
       v2.element = v1.element;
      return;
    }
    
    // we perform the cleanup if and only if the types are different (which avoids one allocation)
    if(v1.element_type.c_type != v2.element_type.c_type)
      cleanup_variant(v2);
        
    switch(v1.element_type.c_type)
    {
      case type::c_vector:
        variant_field_from_type< std::vector<s_any_type> >(v2) = variant_field_from_type< std::vector<s_any_type> >( v1 );     break;
        
      default:
        YAYI_THROW("Undefined type " + string_type(v1.element_type));                     break;
    }
    v2.element_type = v1.element_type;    
  }
  
   
  
  inline void clone_variant_init(const variant& v1, variant& v2) {
    // Simple cases before
    if(v1.element_type.c_type == type::c_scalar) {
      v2.element = v1.element;
      v2.element_type = v1.element_type;      
      return;
    }
    
    switch(v1.element_type.c_type)
    {
      case type::c_vector:
        variant_field_from_type< std::vector<s_any_type> >(v2) = variant_field_from_type< std::vector<s_any_type> >( v1 );     break;
        
      default:
        YAYI_THROW("Undefined type " + string_type(v1.element_type));                     break;
    }
    
    v2.element_type = v1.element_type;
  }
  
  
  
  
  /*!@brief Assign a variant from the provided variable
   *
   * The variant is given the value and the type of the input value u
   */
  template <class U>  inline typename boost::enable_if< type_support<U> >::type 
    transform_variant_utility(const U& u, variant &v)
  {
    v.element_type.c_type = type_support<U>::compound;
    v.element_type.s_type = type_support<U>::scalar;
    variant_field_from_type<U>(v) = u;
  }
  
  template <class U>  inline typename boost::enable_if< type_support< U > >::type 
    transform_variant_utility(const std::vector<U> &u, variant &v) 
  {
    if(!variants_same_type(v, type::s_undefined, type::c_vector)) {
      if(v.element_type.c_type != type::c_scalar)
        cleanup_variant(v);
      v.element_type.c_type = type::c_vector;
      v.element_type.s_type = type::s_undefined;
      v.element.vector_value = new std::vector<variant>();            
    }
    else {
      v.element.vector_value->clear();
    }

    // TODO : faire un constructeur par recopie + assignation pour variant
    for(typename std::vector<U>::size_type i = 0, j = u.size(); i < j; i++)
    {
      (*v.element.vector_value).push_back(u[i]);
    }
  }
  
  
#if 0
  template <class U>  typename boost::disable_if< type_support<U> >::type transform_variant_utility(U, variant &)
  { 
#ifdef YAYI_COMPILER_ERROR_FOR_UNSUPPORTED_TYPES__
    BOOST_STATIC_ASSERT(false);
#endif
    YAYI_THROW(std::string("Unsupported type for variant conversion : ") + typeof(U).name());
  }
#endif
  
  
  
  
  
  template <class U> variant::operator U() const
  {
    if(element_type.c_type != type::c_scalar)
      YAYI_THROW("The containted type is not scalar !: contained type is " + string_type(this->element_type));
    
    // Any scalar type should be convertible to U
    BOOST_STATIC_ASSERT((is_convertible<yaUINT32, U>::value));
    
    switch(element_type.s_type)
    {
      case type::s_bool:
        return static_cast<U>(  variant_field_from_type<yaBool>     (*this) );
      case type::s_ui8:
        return static_cast<U>(  variant_field_from_type<yaUINT8>    (*this) );
      case type::s_i8:
        return static_cast<U>(  variant_field_from_type<yaINT8>     (*this) );
      case type::s_ui16:
        return static_cast<U>(  variant_field_from_type<yaUINT16>   (*this) );
      case type::s_i16:
        return static_cast<U>(  variant_field_from_type<yaINT16>    (*this) );
      case type::s_ui32:
        return static_cast<U>(  variant_field_from_type<yaUINT32>   (*this) );
      case type::s_i32:
        return static_cast<U>(  variant_field_from_type<yaINT32>    (*this) );
      case type::s_ui64:
        return static_cast<U>(  variant_field_from_type<yaUINT64>   (*this) );
      case type::s_i64:
        return static_cast<U>(  variant_field_from_type<yaINT64>    (*this) );
      case type::s_float:
        return static_cast<U>(  variant_field_from_type<yaF_simple> (*this) );
      case type::s_double:
        return static_cast<U>(  variant_field_from_type<yaF_double> (*this) );
      default:
        YAYI_THROW("Undefined type " + string_type(element_type));
    }
  }
  
  
  template <class U> variant::operator std::complex<U>() const
  {
    typedef std::complex<U> return_type;
    if(element_type.c_type != type::c_complex && element_type.c_type != type::c_scalar)
      YAYI_THROW("The contained type is not complex nor scalar ! (contained type is " + string_type(this->element_type) + ")");
    
    switch(element_type.s_type)
    {
      case type::s_float:
        return static_cast<return_type>(  variant_field_from_type< std::complex<yaF_simple> >(*this));
      case type::s_double:
        return static_cast<return_type>(  variant_field_from_type< std::complex<yaF_double> >(*this));
      default:
        YAYI_THROW("Unsupported type " + int_to_string(element_type.s_type));
    }
  }
  
  
  template <class U> variant::operator std::vector<U>() const
  {
    if(element_type.c_type != type::c_vector)
      YAYI_THROW("Unable to transform a type " + string_type(element_type) + std::string(" into a vector type"));
    
    DEBUG_ASSERT(element.vector_value != 0, "The vector has not been initialized");
    std::vector<U> out;
    for(typename std::vector<s_any_type>::size_type i = 0, j = element.vector_value->size(); i < j; i++)
    {
      out.push_back( (*element.vector_value)[i] );
    }
    return out;
  }
  
  
  
  
  
  template <class U, class dimension> variant::operator yayi::s_compound_pixel_t<U, dimension>() const
  {
    typedef s_compound_pixel_t<U, dimension> return_type;
    if(element_type.c_type != type_description::type_support<return_type>::compound)
      YAYI_THROW("The contained type is not a pixel type (" + type_description::type_support<return_type>::name() + ")! contained type is " + string_type(this->element_type));
    
    switch(element_type.s_type)
    {
      case type::s_ui8:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaUINT8, dimension> >    (*this) );
      case type::s_i8:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaINT8, dimension> >     (*this) );
      case type::s_ui16:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaUINT16, dimension> >   (*this) );
      case type::s_i16:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaINT16, dimension> >    (*this) );
      case type::s_ui32:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaUINT32, dimension> >   (*this) );
      case type::s_i32:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaINT32, dimension> >    (*this) );
      case type::s_ui64:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaUINT64, dimension> >   (*this) );
      case type::s_i64:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaINT64, dimension> >    (*this) );
      case type::s_float:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaF_simple, dimension> > (*this) );
      case type::s_double:
        return static_cast<return_type>(  variant_field_from_type<s_compound_pixel_t<yaF_double, dimension> > (*this) );
      default:
        YAYI_THROW("Undefined type " + string_type(element_type));
    }
  }
  
  
  namespace dispatcher {
  
  
  }
  
  

  
  

}

#endif

