/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2009
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */


#include <Yayi/core/yayiCommonPython/include/common_python.hpp>
#include <Yayi/core/yayiCommon/common_types.hpp>
#include <boost/python/enum.hpp>
#include <Yayi/core/yayiCommon/common_variant.hpp>

using namespace yayi;
struct custom_variant_to_python
{

  static PyObject* from_variant_to_python(yayi::variant const& s) {
    
    using yayi::type;

    if(s.element_type.c_type == type::c_scalar) {
      switch(s.element_type.s_type) {
      case type::s_ui8: 
      case type::s_ui16: 
      case type::s_ui32: 
        return PyLong_FromUnsignedLong(static_cast<yaUINT32>(s));
        
      case type::s_ui64: 
        return PyLong_FromUnsignedLongLong(static_cast<yaUINT64>(s));
      
      case type::s_i8: 
      case type::s_i16: 
      case type::s_i32:
        return PyLong_FromLong(static_cast<yaINT32>(s));
      
      case type::s_i64:
        return PyLong_FromLongLong(static_cast<yaINT64>(s));
        
      case type::s_float: 
        return PyFloat_FromDouble(static_cast<yaF_simple>(s));
        
      case type::s_double:
        return PyFloat_FromDouble(static_cast<yaF_double>(s));
      
      case type::s_bool:
      {
        if(variant_field_from_type<yaBool>(s))
          Py_RETURN_TRUE;
        else 
          Py_RETURN_FALSE;
      }
        
      case type::s_string:
        return PyString_FromString(variant_field_from_type<string_type>(s).c_str());
      
      case type::s_wstring:
      {
        const wide_string_type& str_ = variant_field_from_type<wide_string_type>(s);
        return PyUnicode_FromWideChar(str_.c_str(), str_.size());
      }
        
      default:
        YAYI_THROW("Unsupported scalar type");
      }
    
    }
    else if(s.element_type.c_type == type::c_vector) {
      const std::vector<s_any_type>& vect = variant_field_from_type< std::vector<s_any_type> >(s);
      PyObject* out_t = PyTuple_New(vect.size());
      if(out_t == NULL)
        return 0;
      
      std::vector<s_any_type>::size_type i = 0, j = vect.size();
      for(; i < j; i++) {
        PyObject *p = 0;
        try {
          p = from_variant_to_python(vect[i]);
          if(p == NULL || PyTuple_SET_ITEM(out_t, i, p) != 0) {
            // Raffi : pas besoin normalement ??
            //for(std::vector<s_any_type>::size_type k = 0; k < i; k++)
            //  Py_XDECREF(PyTuple_GetItem(out_t, k));
            
            Py_DECREF(out_t);
            return 0;
          }
          Py_DECREF(p);
        } catch(yayi::errors::yaException &e) {
          Py_XDECREF(p);
          Py_DECREF(out_t);
          return 0;          
        }
      }
      return out_t;
    }
    
    YAYI_THROW("Unsupported compound type");
    
    
    return 0;
  }



  static PyObject* convert(yayi::variant const& s)
  {
    return from_variant_to_python(s);
  }
};


struct custom_variant_from_python
{
  custom_variant_from_python()
  {
    boost::python::converter::registry::push_back(
      &convertible,
      &construct,
      boost::python::type_id<yayi::variant>());
  }
  
  static yayi::variant* do_convert(PyObject* obj_ptr, void * pos)
  {
    static const long int l_max = PyInt_GetMax();
    
    if(PyBool_Check(obj_ptr)) {
      bool b = obj_ptr == Py_True;
      return new (pos) variant(b);
    }
      
    else if(PyLong_CheckExact(obj_ptr)) {
      yaINT32 l = PyLong_AsLong(obj_ptr);
      if(l == -1) {
        PyObject * po = PyErr_Occurred();
        if(po != NULL && PyErr_ExceptionMatches(PyExc_OverflowError)) {
          yaINT64 ll = PyLong_AsLongLong(obj_ptr);
          return new (pos) variant(ll);
        }
      }
      return new (pos) variant(l);
    }
    
    else if(PyInt_CheckExact(obj_ptr)) {
      yaINT32 l = PyInt_AsLong(obj_ptr);
      if(l == -1) {
        PyObject * po = PyErr_Occurred();
        if(po != NULL && PyErr_ExceptionMatches(PyExc_OverflowError)) {
          yaINT64 ll = PyLong_AsLongLong(obj_ptr);
          return new (pos) variant(ll);
        }
      }
      return new (pos) variant(l);
    }
#if 0
    else if(PyFloat_CheckExact(obj_ptr))// / PyFloat_AsDouble() / PyFloat_AsDOUBLE() 
      return obj_ptr;
    
    else if(PyString_CheckExact(obj_ptr))// / PyString_Check() / PyString_AsString() 
      return obj_ptr;
    
    else if(PyUnicode_CheckExact(obj_ptr))// / Py_UNICODE* PyUnicode_AS_UNICODE(PyObject *o) 
      return obj_ptr;
    
    else if(PyTuple_CheckExact(obj_ptr))
    {
      for(int i = 0, j = PyTuple_Size(obj_ptr); i < j; i++)
      {
        if(convertible(PyTuple_GetItem(obj_ptr, i)) == 0)
          return 0;
      }
      return obj_ptr;
    }
    
    else if(PyList_CheckExact(obj_ptr))
    {
      for(int i = 0, j = PyList_Size(obj_ptr); i < j; i++)
      {
        if(convertible(PyList_GetItem(obj_ptr, i)) == 0)
          return 0;
      }
      return obj_ptr;
    }
    
    else if(PyDict_CheckExact(obj_ptr))
    {
      PyObject *key = 0, *value = 0;
      int pos = 0;

      while (PyDict_Next(obj_ptr, &pos, &key, &value)) {
        // support only for key of int type and value convertible to variant
        if(!PyInt_CheckExact(key))
          return 0;
        if(convertible(value) == 0)
          return 0;
      }
      return obj_ptr;
    }
#endif
    
    return 0;
  }

  
  
  

  static void* convertible(PyObject* obj_ptr)
  {
    if(obj_ptr == 0)
      return 0;
    //if (!PyString_Check(obj_ptr)) return 0;
    else if(PyBool_Check(obj_ptr))
      return obj_ptr;
      
    else if(PyLong_CheckExact(obj_ptr))
      return obj_ptr;
      
    else if(PyInt_CheckExact(obj_ptr))
      return obj_ptr;
    
    else if(PyFloat_CheckExact(obj_ptr))// / PyFloat_AsDouble() / PyFloat_AsDOUBLE() 
      return obj_ptr;
    
    else if(PyString_CheckExact(obj_ptr))// / PyString_Check() / PyString_AsString() 
      return obj_ptr;
    
    else if(PyUnicode_CheckExact(obj_ptr))// / Py_UNICODE* PyUnicode_AS_UNICODE(PyObject *o) 
      return obj_ptr;
    
    else if(PyTuple_CheckExact(obj_ptr))
    {
      for(int i = 0, j = PyTuple_Size(obj_ptr); i < j; i++)
      {
        if(convertible(PyTuple_GetItem(obj_ptr, i)) == 0)
          return 0;
      }
      return obj_ptr;
    }
    
    else if(PyList_CheckExact(obj_ptr))
    {
      for(int i = 0, j = PyList_Size(obj_ptr); i < j; i++)
      {
        if(convertible(PyList_GetItem(obj_ptr, i)) == 0)
          return 0;
      }
      return obj_ptr;
    }
    
    else if(PyDict_CheckExact(obj_ptr))
    {
      PyObject *key = 0, *value = 0;
      Py_ssize_t pos = 0;

      while (PyDict_Next(obj_ptr, &pos, &key, &value)) {
        // support only for key of int type and value convertible to variant
        if(!PyInt_CheckExact(key))
          return 0;
        if(convertible(value) == 0)
          return 0;
      }
      return obj_ptr;
    }
    
    
    return 0;
  }

  static void construct(
    PyObject* obj_ptr,
    boost::python::converter::rvalue_from_python_stage1_data* data)
  {
    void* storage = ((boost::python::converter::rvalue_from_python_storage<yayi::variant>*)data)->storage.bytes;
    variant *var = do_convert(obj_ptr, storage);
    if (var == 0) boost::python::throw_error_already_set();
    data->convertible = storage;
  }
};


void declare_variants()
{
  boost::python::to_python_converter<yayi::variant, custom_variant_to_python>();
  custom_variant_from_python();
}





