/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2009
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_APPLY_TO_IMAGE_HPP__
#define YAYI_APPLY_TO_IMAGE_HPP__


/*!@file
 * Contains necessary function to iterator operators over images
 *
 * @author Raffi Enficiaud
 */

#include <Yayi/core/yayiImageCore/include/yayiImageUtilities_T.hpp>
#include <Yayi/core/yayiCommon/include/common_string_utilities.hpp>


#include <boost/function_types/function_type.hpp>
#include <boost/function_types/result_type.hpp>
#include <boost/function_types/parameter_types.hpp>
#include <boost/function_types/function_arity.hpp>
#include <boost/function_types/is_member_pointer.hpp>


#include <boost/static_assert.hpp>
#include <boost/mpl/has_xxx.hpp>
#include <boost/utility/result_of.hpp>
#include <boost/function_types/result_type.hpp>



namespace yayi
{

  /*! Strategy to be used for requesting the iterators for more than one image algorithms
   *
   */
  typedef enum e_point_processing_iterator_strategy
  {
    eis_all_different,                      //!< every iterator has to be iterated
    eis_window_one,                         //!< images are the same, one window iterator can be shared
    eis_no_window_one,                      //!< images are the same, one plain iterator can be shared
    eis_window_maximal,                     //!< the iterator over the whole image can be used
    eis_window_maximal_same_offset,         //!< the iterator over the whole image can be used, one iterator can be used if the algorithm shares the offset
    eis_window_maximal_same_pointer,        //!< the iterator over the whole image can be used, one iterator can be used if the algorithm shares the pointer difference
    eis_offset_compatibles,                 //!< the offset from one iterator can be used for accessing the pixel to another image instead of using its iterator
    eis_pointer_compatibles                 //!< the images are offset compatible and have the same pointer arithmetic. 
  } point_processing_iterator_strategy;



  typedef enum e_iterator_choice_strategy
  {
    eic_non_windowed,                       //!< used to indicate that the non windowed iterator should be extracted from the image
    eic_windowed                            //!< not used
  } iterator_choice_strategy;


  //! This tag is for extracting the block version of the iterator
  struct iterator_choice_strategy_non_windowed_tag  {};
  //! This tag is for extracting the windowed version of the iterator
  struct iterator_choice_strategy_windowed_tag      {};


  template <class image>
  iterator_choice_strategy iterator_extractor(const image& im)
  {
    return eic_non_windowed;
  }
  
  template <class image1, class image2>
  iterator_choice_strategy iterator_extractor(const image1& im1, const image2& im2)
  {
    return eic_non_windowed;
  }
  
  template <class image1, class image2, class image3>
  iterator_choice_strategy iterator_extractor(const image1& im1, const image2& im2, const image3& im3)
  {
    return eic_non_windowed;
  }      
  
  
  
  
  
  namespace ns_operator_tag {
  
    /*! The operator tags define some properties of the operators
     *  These properties would allow optimization at run time by splitting (or not) the pixel set into smaller ones,
     *  distributed on different computational units (threads, cpu or machines).
     */
     
    struct operator_no_tag;
    
    
    /*! This tag means that the underlying operator is insensitive to the order of the points on which it is applied to
     *  Example:
     *  \[T(a_1, a_2, ..., a_n) = T({a_i}_{i \in \N^*_n}) = T(a_2, a_1, ..., a_n) = ...\]
     */
    struct operator_commutative {};

    /*! The used operator is a "partition set morphism". This means
     *  that the processing can be split into smaller processings, for instance (T stands for the operator, + is the operator 
     *  in the destination monoid):
     *  
     *  \[\forall j \in \N^*_n, T(a_1, a_2, ..., a_n) = T(a_1, ..., a_j) + T(a_{j+1}, ..., a_n)\]
     */
    struct operator_partition_set_morphism;
  
  
#if 0
    template< typename T > struct trait_msvc_sfinae_helper {
      typedef void type;
    };

    template< typename T, typename U = void > 
    struct trait_impl_
    { 
        typedef boost::mpl::bool_<false> type; 
    }; 

    template< typename T > 
    struct trait_impl_<T, typename trait_msvc_sfinae_helper< typename T::operator_tag >::type > 
    { 
        typedef boost::mpl::bool_<true> type;
    }; 


    template< typename T, typename fallback_ = boost::mpl::bool_<false> >
    struct s_has_tag2 : public trait_impl_<T> 
    { 
    }; 
#endif
  
    BOOST_MPL_HAS_XXX_TRAIT_NAMED_DEF(has_operator_tag, operator_tag, false);

  
  }






  /*!@brief Return the best strategy from the runtime informations over the images for using the iterators and the pixel accesses.
   *
   * @author Raffi Enficiaud
   */
  template <class image1, class image2>
    point_processing_iterator_strategy iterator_strategy(const image1& im1, const image2& im2)
  {
    if(are_images_same(im1, im2))
    {
      return eis_window_maximal_same_pointer;
    }

    if(are_same_geometry(im1, im2))
    {
      // same geometry : if the windows are mxiaml, only one iterator can be used
      if(are_images_same_type(im1, im2))
        return eis_pointer_compatibles;
      else
        return eis_window_maximal_same_offset;
    }
    
    // TODO : add other type of optimizations
    return eis_all_different;
  }
   
  template <class image1, class image2, class image3>
    point_processing_iterator_strategy iterator_strategy(const image1& im1, const image2& im2, const image3& im3)
  {
    if(are_images_same(im1, im2) && are_images_same(im1, im3))
    {
      return eis_window_maximal_same_pointer;
    }

    if(are_same_geometry(im1, im2) && are_same_geometry(im1, im3))
    {
      // same geometry : if the windows are mxiaml, only one iterator can be used
      if(are_images_same_type(im1, im2) && are_images_same_type(im1, im3))
        return eis_pointer_compatibles;
      else
        return eis_window_maximal_same_offset;
    }

    // TODO : add other type of optimizations
    return eis_all_different;
  }

  namespace {


    //! Utility for extracting the correct range of iterators from images
    template <class> 
    struct s_iterator_extractor;

    //! Extracts the correct range of non-windowed iterators from the provided image
    template <> 
    struct s_iterator_extractor<iterator_choice_strategy_non_windowed_tag>
    {
      typedef s_iterator_extractor<iterator_choice_strategy_non_windowed_tag> this_type;
      template <class T> struct results;

      template <class self_, class image> struct results<self_(image&)> {
        typedef typename boost::mpl::if_<
          boost::is_const<image>, 
          typename image::const_iterator, 
          typename image::iterator>::type iterator_type;
        typedef std::pair<iterator_type, iterator_type>   type;
      };
      
      // the image is either const or not const, the compiler knows. The only 
      // important fact is the return type, the called methods are the same.
      template <class image>
      typename results<this_type(image&)>::type operator()(image& im) const
      {
        return make_pair(im.begin_block(), im.end_block());
      }
    };

    //! Extracts the correct range of windowed iterators from the provided image
    template <> 
    struct s_iterator_extractor<iterator_choice_strategy_windowed_tag>
    {
      typedef s_iterator_extractor<iterator_choice_strategy_windowed_tag> this_type;
      template <class T> struct results;

      template <class self_, class image> struct results<self_(image&)> {
        typedef typename boost::mpl::if_<
          boost::is_const<image>, 
          typename image::const_window_iterator, 
          typename image::window_iterator>::type iterator_type;

        typedef std::pair<iterator_type, iterator_type>   type;
      };

      // the image is either const or not const, the compiler knows. The only 
      // important fact is the return type, the called methods are the same.
      template <class image> 
      typename results<this_type(image&)>::type operator()(image& im) const
      {
        return make_pair(im.begin_window(), im.end_window());
      }
    };

  }








  struct iterators_independant_tag            {};     //!< The input iterators are independant
  struct iterators_same_number_of_points_tag  {};     //!< The input iterators are iterating over the same number of points (useful ?)
  struct iterators_same_offset_tag            {};     //!< The iterators have the same offset
  struct iterators_same_offset_and_same_images_tag {};     //!< Same as iterators_same_offset_tag but the images are also the same

  struct operator_type_zero_ary_no_return     {};     //!< No input, no output (eg. a counter)
  struct operator_type_zero_ary               {};     //!< No input, but an output (eg. a generator)
  
  struct operator_type_unary                  {};     //!< An input and an output (eg. a transformer)
  struct operator_type_unary_no_return        {};     //!< An input but not output (eg. an applyier)
  
  struct operator_type_binary                 {};
  struct operator_type_binary_no_return       {};



  BOOST_MPL_HAS_XXX_TRAIT_NAMED_DEF(operator_has_result_member_tag, result, false);





  /*!@brief Utility structure for extracting the type of the supplied operator
   *
   * @author Raffi Enficiaud
   */
  struct s_extract_operator_type
  {
    template <class op_> struct result;
    
    // Revoir l'utilité de cette fonction ici. Grosso modo: on ne cherche pas à savoir le nombre de paramètres avec lequel
    // on appelle cette fonction, mais plutôt de connaitre le type de la fonction.
    template <class op_> 
    struct result< op_() > {
      /*typedef typename mpl::if_<
        operator_has_result_member_tag<op_>,
        typename boost::result_of< op_() >::type,
        typename result_type< &(op_::operator()()) >::type>::type result_type;*/
      typedef typename boost::result_of< op_() >::type result_type;
      typedef typename mpl::if_<
        typename is_void<result_type>::type, 
        operator_type_zero_ary_no_return, 
        operator_type_zero_ary>::type type;
      //typedef operator_type_zero_ary_no_return type; 
    };

    template <class op_, class T> 
    struct result< op_(T) > {
      typedef typename boost::result_of< op_(T) >::type result_type;
      typedef typename mpl::if_<
        typename is_void<result_type>::type, 
        operator_type_unary_no_return, 
        operator_type_unary>::type type;
      //typedef operator_type_unary type; 
    };

    template <class op_, class T, class U> 
    struct result< op_(T, U)> {
      typedef typename boost::result_of< op_(T, U) >::type result_type;
      typedef typename mpl::if_<
        typename is_void<result_type>::type, 
        operator_type_binary_no_return, 
        operator_type_binary>::type type;
    };

  };
  
  typedef s_extract_operator_type operator_traits;



  //! Declaration of the iterator application utllity structure
  template <class iterators_interaction_t, class operator_type_t>
  struct s_apply_op_range;
  
  //! Specializing of s_apply_op_range for zero ary operators (no return)
  template <>
  struct s_apply_op_range<iterators_independant_tag, operator_type_zero_ary_no_return>
  {
    template <class op_, class iter1, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, image1&)
    {
      // to do: just count the number of element inside the iterator range would be enough there
      for(; it1 != it1e; ++it1){
        op();
      }
      return yaRC_ok;
    }
  };
  
  //! Specializing of s_apply_op_range for zero ary operators (with return)
  template <>
  struct s_apply_op_range<iterators_independant_tag, operator_type_zero_ary>
  {
    template <class op_, class iter1, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, image1&)
    {
      for(; it1 != it1e; ++it1){
        *it1 = op();
      }
      return yaRC_ok;
    }
  };


  //! Specializing of s_apply_op_range for independant iterators and unary operators
  template <class it_strategy /*= iterators_independant_tag*/>
  struct s_apply_op_range<it_strategy, operator_type_unary>
  {
    template <class op_, class iter1, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, image1&)
    {
      for(; it1 != it1e; ++it1){
        op(*it1);
      }
      return yaRC_ok;
    }
    
    
    //! Unary operation with return value
    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2 it2, const iter2 it2e, image1&, image2&)
    {
      // todo : some cases where im1 == im2 && it1 == it2 && it1e == it2e should be optimized
      for(; it1 != it1e && it2 != it2e; ++it1, ++it2){
        *it2 = op(*it1);
      }
      return yaRC_ok;
    }
  };

  //! Specializing of s_apply_op_range for iterators sharing the same offset configuration
  template <>
  struct s_apply_op_range<iterators_same_offset_tag, operator_type_unary>
  {
    //! Unary operation with return value
    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2, const iter2, image1& im1, image2& im2)
    {
      // todo : some cases where im1 == im2 && it1 == it2 && it1e == it2e should be optimized
      for(; it1 != it1e; ++it1){
        const offset o = it1.Offset();
        im2.pixel(o) = op(im1.pixel(o));
      }
      return yaRC_ok;
    }
  };

  //! Specializing of s_apply_op_range for iterators sharing the same offset configuration
  template <>
  struct s_apply_op_range<iterators_same_offset_and_same_images_tag, operator_type_unary>
  {
    //! Unary operation with return value
    template <class op_, class iter1, class iter2, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2, const iter2, image1& im1, image1&)
    {
      typedef typename mpl::if_<
        is_const<image1>,
        typename image1::const_reference,
        typename image1::reference>::type ref_t;    
    
      for(; it1 != it1e; ++it1){
        ref_t p = *it1;
        p = op(p);
      }
      return yaRC_ok;
    }

    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_&, iter1, const iter1, iter2, const iter2, image1&, image2&)
    {
      YAYI_THROW("This function should never be called : the program is ill-formed !");
    }
  };
  
  
  



  
  //! Specializing of s_apply_op_range for independant iterators and unary operators without return value
  template <>
  struct s_apply_op_range<iterators_independant_tag, operator_type_unary_no_return>
  {
    template <class op_, class iter1, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, image1&)
    {
      for(; it1 != it1e; ++it1){
        op(*it1);
      }
      return yaRC_ok;
    }

  };
  
  
  
  //! Binary operator helper
  //! The images structure is different, and no factorisation of the iterator can be performed.
  //! We iterate over the two images, until one of them ends. 
  template <>
  struct s_apply_op_range<iterators_independant_tag, operator_type_binary_no_return>
  {
    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2 it2, const iter2 it2e, image1&, image2&)
    {
      for(; it1 != it1e && it2 != it2e; ++it1, ++it2){
        op(*it1, *it2);
      }
      return yaRC_ok;
    }
  };


  //! Binary operator helper



  //! The images structure is different, and no factorisation of the iterator can be performed.
  //! We iterate over the two images, until one of them ends. 
  template <>
  struct s_apply_op_range<iterators_independant_tag, operator_type_binary>
  {
    template <class op_, class iter1, class iter2, class iter3, class image1, class image2, class image3>
    yaRC operator()(
      op_& op, 
      iter1 it1, const iter1 it1e, 
      iter2 it2, const iter2 it2e, 
      iter3 it3, const iter3 it3e, 
      image1&, image2&, image3&)
    {
      for(; it1 != it1e && it2 != it2e && it3 != it3e; ++it1, ++it2, ++it3){
        *it3 = op(*it1, *it2);
      }
      return yaRC_ok;
    }

    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(
      op_& op, 
      iter1 it1, const iter1 it1e, 
      iter2 it2, const iter2 it2e, 
      image1&, image2&)
    {
      for(; it1 != it1e && it2 != it2e; ++it1, ++it2){
        op(*it1, *it2);
      }
      return yaRC_ok;
    }

  };


  //! Images share the same offset structure
  //! It is hence possible to use one iterator instead of two, if the ranges allow to do so.
  template <>
  struct s_apply_op_range<iterators_same_offset_tag, operator_type_binary>
  {
    template <class op_, class iter1, class iter2, class iter3, class image1, class image2, class image3>
    yaRC operator()(
      op_& op, 
      iter1 it1, const iter1 it1e, 
      iter2 , const iter2 , 
      iter3 , const iter3 , 
      image1& im1, image2& im2, image3& im3)
    {
      for(; it1 != it1e; ++it1){
        const offset o = it1.Offset();
        im3.pixel(o) = op(im1.pixel(o), im2.pixel(o));
      }
      return yaRC_ok;
    }  

    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(
      op_& op, 
      iter1 it1, const iter1 it1e, 
      iter2 , const iter2 , 
      image1& im1, image2& im2)
    {
      for(; it1 != it1e; ++it1){
        const offset o = it1.Offset();
        op(im1.pixel(o), im2.pixel(o));
      }
      return yaRC_ok;
    }  

  };
  
    //! Images share the same offset structure
  //! It is hence possible to use one iterator instead of two, if the ranges allow to do so.
  template <>
  struct s_apply_op_range<iterators_same_offset_tag, operator_type_binary_no_return>
  {
    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2, const iter2, image1& im1, image2& im2)
    {
      for(; it1 != it1e; ++it1){
        const offset o = it1.Offset();
        op(im1.pixel(o), im2.pixel(o));
      }
      return yaRC_ok;
    }
  };
  



  //! Specializing of s_apply_op_range for iterators sharing the same images
  template <>
  struct s_apply_op_range<iterators_same_offset_and_same_images_tag, operator_type_binary_no_return>
  {
    //! Unary operation with return value
    template <class op_, class iter1, class iter2, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2, const iter2, image1& im1, image1&)
    {
      typedef typename mpl::if_<
        is_const<image1>,
        typename image1::const_reference,
        typename image1::reference>::type ref_t;
            
      for(; it1 != it1e; ++it1){
        ref_t p = *it1;
        op(p, p);
      }
      return yaRC_ok;
    }

    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_&, iter1, const iter1, iter2, const iter2, image1&, image2&)
    {
      YAYI_THROW("This function should never be called : the program is ill-formed !");
    }
  };

  //! Specializing of s_apply_op_range for iterators sharing the same images
  template <>
  struct s_apply_op_range<iterators_same_offset_and_same_images_tag, operator_type_binary>
  {
    //! Unary operation with return value
    template <class op_, class iter1, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter1, const iter1, image1& im1, image1&)
    {
      typedef typename mpl::if_<
        is_const<image1>,
        typename image1::const_reference,
        typename image1::reference>::type ref_t;
      for(; it1 != it1e; ++it1){
        ref_t p = *it1;
        op(p, p);
      }
      return yaRC_ok;
    }

    template <class op_, class iter1, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter1, const iter1, iter1, const iter1, image1& im1, image1&, image1&)
    {
      typedef typename image1::reference ref_t;
      /*typedef typename mpl::if_<
        is_const<image1>,
        typename image1::const_reference,
        typename image1::reference>::type ref_t;*/
      for(; it1 != it1e; ++it1){
        ref_t p = *it1;
        p = op(p, p);
      }
      return yaRC_ok;
    }


    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_&, iter1, const iter1, iter2, const iter2, image1&, image2&)
    {
      YAYI_THROW("This function should never be called : the program is ill-formed !");
    }

    template <class op_, class iter1, class iter2, class iter3, class image1, class image2, class image3>
    yaRC operator()(op_&, iter1, const iter1, iter2, const iter2, iter3, const iter3, image1&, image2&, image3&)
    {
      YAYI_THROW("This function should never be called : the program is ill-formed !");
    }

  };

  


  /*!@brief Unary operator functor over an image
   *
   * @author Raffi Enficiaud
   */
  struct s_apply_unary_operator
  {

  private:
    template <class iterator_extractor_t, class image_in_out, class op_>
    yaRC apply(image_in_out& im, op_& op) const
    {
      //BOOST_STATIC_ASSERT(is_void<typename op_::result_type>::value);
      typedef s_iterator_extractor<iterator_extractor_t> extractor_type;
      
      extractor_type extractor;
      
      typedef typename extractor_type::template results<extractor_type(image_in_out&)>::type result_type;
      
      const result_type iterators(extractor(im));
      typename result_type::first_type  it = iterators.first;
      typename result_type::second_type const ite = iterators.second;
      
      // test the return of op if it exists ??
      typedef typename s_extract_operator_type::template result< op_(typename result_type::first_type::reference_type) >::type range_type;
      
      
      
      // Here we can add more precision on the type of 
      s_apply_op_range<iterators_independant_tag, range_type> op_apply;
      
      return op_apply(op, it, ite, im);
    }

    // TODO : separate the extractors for each image
    template <class iterator_extractor_t, class image_in, class image_out, class op_>
    yaRC apply(image_in& im, image_out& out, op_& op) const
    {
      //BOOST_STATIC_ASSERT(is_void<typename op_::result_type>::value);
      typedef s_iterator_extractor<iterator_extractor_t> extractor_type;
      
      extractor_type extractor;
      
      typedef typename extractor_type::template results<extractor_type(image_in&)>::type itrange_in_t;
      typedef typename extractor_type::template results<extractor_type(image_out&)>::type itrange_out_t;
      
      itrange_in_t const iterators_in(extractor(im));
      typename itrange_in_t::first_type  it = iterators_in.first;
      typename itrange_in_t::second_type const ite = iterators_in.second;
      
      itrange_out_t const iterators_out(extractor(out));
      typename itrange_out_t::first_type  ito = iterators_out.first;
      typename itrange_out_t::second_type const itoe = iterators_out.second;
      
      // test the return of op if it exists ??
      typedef typename s_extract_operator_type::template result< op_(typename itrange_in_t::first_type::reference_type) >::type range_type;
      
      
      switch(iterator_strategy(im, out))
      {
        case eis_window_maximal_same_pointer:
        {
          s_apply_op_range<iterators_same_offset_and_same_images_tag, range_type> op_apply;
          return op_apply(op, it, ite, ito, itoe, im, out);      
        }
        
        case eis_pointer_compatibles:
        case eis_window_maximal_same_offset:
        case eis_offset_compatibles:
        {
          s_apply_op_range<iterators_same_offset_tag, range_type> op_apply;
          return op_apply(op, it, ite, ito, itoe, im, out);
        }
        
        default: 
        {
          s_apply_op_range<iterators_independant_tag, range_type> op_apply;
          return op_apply(op, it, ite, ito, itoe, im, out);
        }
    
      }  
      
    }
    



  public:
    template <class image_in_out, class op_>
      yaRC operator()(image_in_out& im, op_& op) const
    {
      switch(iterator_extractor(im))
      {
      case eic_non_windowed:
        {
          return apply<iterator_choice_strategy_non_windowed_tag>(im, op);
        }
      /*case eic_windowed:
        {
          return apply<image_in_out, iterator_choice_strategy_windowed_tag>(im, op);
        }*/
      default:
        YAYI_THROW("Unsupported case " + any_to_string(iterator_extractor(im)));
      }
    }

    template <class image_in, class image_out, class op_>
      yaRC operator()(const image_in& im, image_out& out, op_& op) const
    {
      switch(iterator_extractor(im))
      {
      case eic_non_windowed:
        {
          return apply<iterator_choice_strategy_non_windowed_tag>(im, out, op);
        }
      /*case eic_windowed:
        {
          return apply<image_in_out, iterator_choice_strategy_windowed_tag>(im, op);
        }*/
      default:
        YAYI_THROW("Unsupported case " + any_to_string(iterator_extractor(im)));
      }
    }



  };





  
  
  /*! Utility function for operations involving two images
   *
   */
  struct s_apply_binary_operator
  {

  private:
    template <class iterator_extractor1_t, class iterator_extractor2_t, class image1, class image2, class op_>
      yaRC apply(image1& im1, image2& im2, op_& op) const
    {
      //BOOST_STATIC_ASSERT(is_void<typename op_::result_type>::value);
      typedef s_iterator_extractor<iterator_extractor1_t> extractor1_type;
      typedef s_iterator_extractor<iterator_extractor2_t> extractor2_type;
      
      extractor1_type extractor1;
      extractor2_type extractor2;
      
      typedef typename extractor1_type::template results<extractor1_type(image1&)>::type result1_type;
      typedef typename extractor2_type::template results<extractor2_type(image2&)>::type result2_type;
      
      const result1_type iterators1(extractor1(im1));
      const result2_type iterators2(extractor2(im2));
      
      typename result1_type::first_type        it1 = iterators1.first;
      const typename result1_type::second_type it1e= iterators1.second;
      

      typename result2_type::first_type        it2 = iterators2.first;
      const typename result2_type::second_type it2e= iterators2.second;

      // test the return of op if it exists ??
      typedef typename s_extract_operator_type::template result< op_(typename result1_type::first_type::reference_type, typename result2_type::first_type::reference_type) >::type range_type;

      switch(iterator_strategy(im1, im2))
      {
        case eis_window_maximal_same_pointer:
        {
          s_apply_op_range<iterators_same_offset_and_same_images_tag, range_type> op_apply;
          return op_apply(op, it1, it1e, it2, it2e, im1, im2);      
        }
        
        case eis_pointer_compatibles:
        case eis_window_maximal_same_offset:
        case eis_offset_compatibles:
        {
          s_apply_op_range<iterators_same_offset_tag, range_type> op_apply;
          return op_apply(op, it1, it1e, it2, it2e, im1, im2);          
        }
        
        default: 
        {
          s_apply_op_range<iterators_independant_tag, range_type> op_apply;
          return op_apply(op, it1, it1e, it2, it2e, im1, im2);  
        }
    
      }
      
    }

    template <
      class iterator_extractor1_t, class iterator_extractor2_t, class iterator_extractor3_t, 
      class image1, class image2, class image3, 
      class op_>
      yaRC apply(image1& im1, image2& im2, image3& im3, op_& op) const
    {
      //BOOST_STATIC_ASSERT(is_void<typename op_::result_type>::value);
      typedef s_iterator_extractor<iterator_extractor1_t> extractor1_type;
      typedef s_iterator_extractor<iterator_extractor2_t> extractor2_type;
      typedef s_iterator_extractor<iterator_extractor3_t> extractor3_type;
      
      extractor1_type extractor1;
      extractor2_type extractor2;
      extractor3_type extractor3;
            
      typedef typename extractor1_type::template results<extractor1_type(image1&)>::type result1_type;
      typedef typename extractor2_type::template results<extractor2_type(image2&)>::type result2_type;
      typedef typename extractor3_type::template results<extractor3_type(image3&)>::type result3_type;
      
      const result1_type iterators1(extractor1(im1));
      const result2_type iterators2(extractor2(im2));
      const result3_type iterators3(extractor3(im3));
      
      typename result1_type::first_type        it1 = iterators1.first;
      const typename result1_type::second_type it1e= iterators1.second;
      
      typename result2_type::first_type        it2 = iterators2.first;
      const typename result2_type::second_type it2e= iterators2.second;

      typename result3_type::first_type        it3 = iterators3.first;
      const typename result3_type::second_type it3e= iterators3.second;



      // test the return of op if it exists ??
      typedef typename s_extract_operator_type::template result< 
        op_(
          typename result1_type::first_type::reference_type, 
          typename result2_type::first_type::reference_type
          ) >::type range_type;

      switch(iterator_strategy(im1, im2, im3))
      {
        case eis_window_maximal_same_pointer:
        {
          s_apply_op_range<iterators_same_offset_and_same_images_tag, range_type> op_apply;
          return op_apply(op, it1, it1e, it2, it2e, it3, it3e, im1, im2, im3);      
        }
        
        case eis_pointer_compatibles:
        case eis_window_maximal_same_offset:
        case eis_offset_compatibles:
        {
          s_apply_op_range<iterators_same_offset_tag, range_type> op_apply;
          return op_apply(op, it1, it1e, it2, it2e, it3, it3e, im1, im2, im3);          
        }
        default: 
        {
          s_apply_op_range<iterators_independant_tag, range_type> op_apply;
          return op_apply(op, it1, it1e, it2, it2e, it3, it3e, im1, im2, im3);  
        }
    
      }
      
    }



  public:
  
    template <class image_in1_t, class image_in2_t, class op_>
      yaRC operator()(image_in1_t& in1, image_in2_t& in2, op_& op) const
    {
    
      switch(iterator_extractor(in1, in2))
      {
      case eic_non_windowed:
        {
          return apply<iterator_choice_strategy_non_windowed_tag, iterator_choice_strategy_non_windowed_tag>(in1, in2, op);
        }
      default:
        YAYI_THROW("Unsupported case " + any_to_string(iterator_extractor(in1, in2)));
      }
    }    
    
    

      
    template <class image_in1_t, class image_in2_t, class image_out1_t, class op_>
      yaRC operator()(image_in1_t& in1, image_in2_t& in2, image_out1_t& o, op_& op) const
    {
      switch(iterator_extractor(in1, in2, o))
      {
      case eic_non_windowed:
        {
          return apply<
            iterator_choice_strategy_non_windowed_tag, 
            iterator_choice_strategy_non_windowed_tag,
            iterator_choice_strategy_non_windowed_tag>(in1, in2, o, op);
        }
      default:
        YAYI_THROW("Unsupported case " + any_to_string(iterator_extractor(in1, in2, o)));
      }
    }
    
  };


  
  
  
  template <class image_, class op_>
  yaRC Process(image_& im, op_& op) {
    s_apply_op_range<iterators_independant_tag, typename op_::operator_tag> op_process;
    return op_process(op, im.begin(), im.end(), im);
  }

  template <class image1_, class op_, class image2_>
  yaRC Process(image1_& im1, op_& op, image2_& im2) {
    
    
    // ici il faut iterer en fonction des cas runtime
    s_apply_op_range<iterators_independant_tag, typename op_::operator_tag> op_process;

    return op_process(op, im1.begin(), im1.end(), im1);
  }
  
  
  
  
}


#endif /* YAYI_APPLY_TO_IMAGE_HPP__ */

