/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2009
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_RUNTIME_STRUCTURING_ELEMENT_HEXAGON_HPP__
#define YAYI_RUNTIME_STRUCTURING_ELEMENT_HEXAGON_HPP__

/*!@file
 * This file defines the hexagonal structuring element. More precisely, it instanciates 
 * the correct list chooser (alternating on the y axis). 
 */

#include <Yayi/core/yayiStructuringElement/yayiRuntimeStructuringElement.hpp>

namespace yayi { namespace se {


  /*! This helper class implements the centering behaviour concept for hexagonal structuring element
   *  It basically returns two alternating neighbor list along y axis. 
   */
  template <class coordinate_type>
  struct s_alternating_on_y_behaviour : public std::unary_function<coordinate_type, yaUINT8 > 
  {
    typedef std::unary_function<coordinate_type, yaUINT8> parent_type;
    typedef typename parent_type::result_type result_type; // Raffi : this is redundant ...
    
    s_alternating_on_y_behaviour(bool is_transposed_) : 
      //min_(std::numeric_limits<unsigned int>::max()), med_(min_), max_(min_),
      is_transposed(is_transposed_)
    {} 
    
    result_type operator()(const coordinate_type& p) const throw() {
      //return [min_, med_[, or [med_, max_[
      return ((static_cast<yaUINT8>(p[1]) & 1) == 0) ^ is_transposed;
    }

    result_type operator()(const offset o) const throw() {
      //return [min_, med_[, or [med_, max_[
      return ((static_cast<yaUINT8>(o / off_x) & 1) == 0) ^ is_transposed;
    }
    
    template <class se_t, class im_t, class container>
    void init_indexes(const se_t& se, const im_t& im_, container& c) {
      DEBUG_ASSERT(se.size() % 2 == 0, "The number of points is odd (incorrect)");
      c.clear();
      c.push_back(std::make_pair(0, se.size() / 2));
      c.push_back(std::make_pair(se.size() / 2, se.size()));
      off_x = im_.Size()[0];
    }
    const bool  is_transposed;
    offset      off_x;
  };


  /*! Implements the behaviour of an hexagon structuring element.
   *  This SE does not reshape on x shifts, but does reshape on all other type of shifts.
   *  It contains two alternating lists of structuring elements.
   */
  template <
    class coordinate_type,
    class element_t = coordinate_type,    
    class se_specific_ = neighborlist_helper_multiple_list<
                          coordinate_type,
                          s_alternating_on_y_behaviour<coordinate_type>,
                          structuring_element::se_tags::structuring_element_reshape_on_coordinate_no_x>
    >
  struct s_neighborlist_se_hexa_x : 
    public s_neighborlist_se <coordinate_type, element_t, se_specific_>
  {
  public:
    typedef s_neighborlist_se<coordinate_type, element_t, se_specific_> parent_type;
    typedef typename parent_type::storage_type storage_type;
    typedef s_neighborlist_se_hexa_x<coordinate_type, element_t, se_specific_> this_type;
  
    s_neighborlist_se_hexa_x() : parent_type() {}
    
    s_neighborlist_se_hexa_x(const storage_type& v) : parent_type(v, 2) {}
    
    virtual e_structuring_element_subtype GetSubType() const {
      return e_sest_neighborlist_hexa;
    }
    
    //IObject interface
    //! Object type
    virtual type DynamicType() const {
      return type(type::c_structuring_element, type::s_object);
    }

    //! Object description
    virtual string_type Description() const {
      static const string_type s = "Structuring element reshaping along Y";
      return s;
    }

  };


} }

#endif /* YAYI_RUNTIME_STRUCTURING_ELEMENT_HEXAGON_HPP__ */
