/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2009
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_STRUCTURING_ELEMENT_MAIN_HPP__
#define YAYI_STRUCTURING_ELEMENT_MAIN_HPP__





/*@namespace structuring
 *@brief Structuring element structures and elements
 *
 *
 *@author Raffi Enficiaud
 */



#include "Yayi/core/yayiCommon/common_config.hpp"
#include "Yayi/core/yayiImageCore/include/yayiImageCore.hpp"


#ifdef YAYI_EXPORT_SE_
#define YSE_ MODULE_EXPORT
#else
#define YSE_ MODULE_IMPORT
#endif


namespace yayi {

  namespace se {


    typedef enum e_structuring_element_type {
      e_set_runtime,
      e_set_neighborlist,                       //!< Neighborlist structuring element
      e_set_image,                              //!< Image structuring element
      e_set_functional,
      e_set_template,                           //!< Complete template type of structuring element (cannot be used for dynamic definition, eg. for internal use)
      e_set_chain                               //!< Chain of structuring element
    
    } structuring_element_type;


    typedef enum e_structuring_element_subtype {
      e_sest_neighborlist_generic_single,       //!< generic single list structuring element
      e_sest_neighborlist_hexa                  //!< hexagonal reshaping structuring element
    } structuring_element_subtype;


    /*! Interface for structuring elements
     *  @author Raffi Enficiaud
     */
    class IStructuringElement : public IObject {
    public:
      virtual ~IStructuringElement(){}

      //! Returns a new structuring element that is a transposed copy of this one
      virtual IStructuringElement* Transpose() const    = 0;

      //! Returns the runtime type of the structuring element (see @ref structuring_element_type)
      virtual structuring_element_type GetType() const  = 0;
      
      virtual e_structuring_element_subtype GetSubType() const  = 0;
      
      //! Returns the size of the structuring element in number of neighbors (the center, if defined, is included)
      //! @todo: Raffi: maybe should be a property of a special kind of structuring elements only
      virtual unsigned int GetSize() const              = 0;

      //! Returns a copy of this structuring element
      virtual IStructuringElement* Clone() const        = 0;
      
    };
    
    
    
    /*! Interface for the neighborhood
     *  @author Raffi Enficiaud
     */
    class IConstNeighborhood : public IObject {
    public:
      virtual ~IConstNeighborhood(){}
    
      //! The iterator type for the neighborhood (generic image iterator interface)
      typedef IImage                                image_type;
      typedef image_type::coordinate_type           coordinate_type;
      typedef IIterator*                            iterator;
      typedef IConstIterator*                       const_iterator;
      
      //! Centers the Neighborhood at a particular point
      virtual yaRC Center(const coordinate_type&) = 0;
      
      //! Centers the neighborhood at the point defined by the iterator
      virtual yaRC Center(const const_iterator&) = 0;

      //! Centers the neighborhood at the point defined by the iterator
      virtual yaRC Center(const offset) = 0;
           
      /*!This function defines the shift between two successive calls to 
       * the center method. 
       */
      virtual yaRC SetShift(const coordinate_type&) = 0;
      
      /*! Shifts the centers of the neighborhood with an offset defined 
       *  by the call to @ref set_shift.
       */
      virtual yaRC ShiftCenter() = 0;
      
      //! Returns a const iterator to the beginning of the neighborhood
      virtual const_iterator BeginConst() const = 0;
      
      //! Returns a const iterator to the end of the neighborhood
      virtual const_iterator EndConst()   const = 0;      
      
      //! Factory for the structuring element
      static IConstNeighborhood* Create(const IImage&, const IStructuringElement&);
    };


    /*! Interface for the neighborhood (with write accesses)
     *  @author Raffi Enficiaud
     */  
    class INeighborhood : public IConstNeighborhood {
    public:
  
      virtual iterator Begin() const = 0;
      virtual iterator End()   const = 0;      
      
      //! Factory for the structuring element
      static INeighborhood* Create(IImage&, const IStructuringElement&);  
    };
  
  
  
  }
}



#endif

